#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/2 20:02
# @Author  : 兵
# @email    : 1747193328@qq.com
from PySide6.QtGui import QBrush, QColor
from enum import Enum

class ForcesMode(Enum):
    Row="Row"
    Norm="Norm"

class Brushes:
    # 基本颜色刷子
    RedBrush = QBrush(QColor(255, 0, 0))  # 红色
    GreenBrush = QBrush(QColor(0, 255, 0))  # 绿色
    BlueBrush = QBrush(QColor(0, 0, 255))  # 蓝色
    YellowBrush = QBrush(QColor(255, 255, 0))  # 黄色
    TransparentBrush = QBrush(QColor(0, 0, 0,0))  # 黄色



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/20 21:45
# @Author  : 兵
# @email    : 1747193328@qq.com
import json
import os.path
import sys
import numpy as np
from PySide6.QtGui import QColor

from PySide6.QtWidgets import QApplication, QMainWindow, QVBoxLayout, QWidget
import pyqtgraph as pg
import pyqtgraph.opengl as gl

from NepTrainKit.core import MessageManager
from NepTrainKit import module_path

class StructurePlotWidget(gl.GLViewWidget):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

        # 创建 PyQtGraph 的 3D 窗口
        self.setBackgroundColor('w')
        self.setCameraPosition(distance=30, elevation=30, azimuth=30)
        with open(os.path.join(module_path,"Config/ptable.json"), "r",encoding="utf-8") as f:
            self.table_info=json.loads(f.read())
        # self.setAntialiasing(True)

    def show_lattice(self,lattice_matrix):


        # 定义晶格顶点
        origin = np.array([0.0, 0.0, 0.0])
        a1 = lattice_matrix[0]
        a2 = lattice_matrix[1]
        a3 = lattice_matrix[2]

        vertices = np.array([
            origin,  # (0, 0, 0)
            a1,  # a1
            a2,  # a2
            a3,  # a3
            a1 + a2,  # a1 + a2
            a1 + a3,  # a1 + a3
            a2 + a3,  # a2 + a3
            a1 + a2 + a3  # a1 + a2 + a3
        ])

        # 定义线条连接的顶点索引
        edges = [
            [0, 1], [0, 2], [0, 3],  # 从原点到三个基矢量的线
            [1, 4], [1, 5],  # a1 与 a1 + a2, a1 + a3
            [2, 4], [2, 6],  # a2 与 a1 + a2, a2 + a3
            [3, 5], [3, 6],  # a3 与 a1 + a3, a2 + a3
            [4, 7], [5, 7], [6, 7]  # 顶部的三个线
        ]

        # 将顶点坐标转换为线段的起点和终点
        lines = []
        for edge in edges:
            lines.append(vertices[edge])

        lines = np.array(lines).reshape(-1, 3)

        # 绘制晶格线条
        lattice_lines = gl.GLLinePlotItem(pos=lines, color=(0,0,0,1),
                                          width=1.5, mode='lines',glOptions="translucent",
                                          antialias=True

                                          )
        # lattice_lines.color=(0,1,0,1)
        center = lattice_matrix.sum(axis=0) / 2
        self.opts['center'] = pg.Vector(center[0], center[1], center[2])
        self.addItem(lattice_lines)


    def show_elem(self,numbers,postions):


        # colors=np.array([list(QColor(self.table_info[str(n)]["color"]).getRgbF()) for n in numbers ])
        # print(colors)
        # sizes=np.array([self.table_info[str(n)]["radii"]//5 for n in numbers])
        # colors[:,3]=0.3
        for n,p in zip(numbers,postions):
            color=QColor(self.table_info[str(n)]["color"]).getRgbF()
            size=self.table_info[str(n)]["radii"]/100

            sphere = gl.MeshData.sphere(rows=20, cols=20,radius=size)
            m = gl.GLMeshItem(meshdata=sphere, smooth=True, shader='shaded', color=color)
            self.addItem(m)

            m.translate(p[0], p[1], p[2])  # 设置球体的位置


        # items=gl.GLScatterPlotItem(pos=postions, color=colors ,size=sizes, pxMode=False,
    # glOptions='opaque')
        # items.setGLOptions('opaque')  # 确保散点不透明




    def show_atoms(self,atoms):

        self.clear()

        self.show_lattice(atoms.cell )
        self.show_elem(atoms.numbers,atoms.positions)
if __name__ == '__main__':
    app = QApplication([])
    view = AtomsPlotWidget()
    view.show()
    QApplication.instance().exec_()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/17 13:03
# @Author  : 兵
# @email    : 1747193328@qq.com
import time
from abc import abstractmethod

import numpy as np
from PySide6.QtCore import Signal
from pyqtgraph import GraphicsLayoutWidget, mkPen, ScatterPlotItem




class PlotBase:



    def __init__(self):
        self.current_plot=self

    @abstractmethod
    def select(self,*args,**kwargs):
        raise NotImplementedError
    @abstractmethod
    def delete(self,*args,**kwargs):
        raise NotImplementedError

    def select_point_from_polygon(self,polygon_x,polygon_y):
        pass
    @staticmethod
    def is_point_in_polygon(points, polygon):
        """
        判断多个点是否在多边形内
        :param points: (N, 2) 的数组，表示 N 个点的坐标
        :param polygon: (M, 2) 的数组，表示多边形的顶点坐标
        :return: (N,) 的布尔数组，表示每个点是否在多边形内
        """
        n = len(polygon)
        inside = np.zeros(len(points), dtype=bool)

        px, py = points[:, 0], points[:, 1]
        p1x, p1y = polygon[0]

        for i in range(n + 1):
            p2x, p2y = polygon[i % n]
            mask = ((py > np.minimum(p1y, p2y)) &
                    (py <= np.maximum(p1y, p2y)) &
                    (px <= np.maximum(p1x, p2x)) &
                    (p1y != p2y))
            xinters = (py[mask] - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
            inside[mask] ^= (px[mask] <= xinters)
            p1x, p1y = p2x, p2y

        return inside



class LayoutPlotBase(PlotBase):
    def __init__(self,row=None,col=None,**kwargs):
        super().__init__()
        self.current_plot=None
        self.current_view_box=None
        self.tool_bar=None
        self.axes_list=[]
        self.width = 3
        self.height = 3
        if row is not None and col is not None:
            self.subplot(row,col)

    def subplot(self,row,col):

        self.axes_list.clear()
        self.rows=row
        self.cols=col
        self.clear()

        for r in range(5):
            if r==0:
                plot = self.addPlot(row=r, col=0,colspan=3 )
            else:

                plot = self.addPlot(row=1, col=r-1  )
            plot.getViewBox().mouseDoubleClickEvent = self.on_click(plot)
            plot.getViewBox().setMouseEnabled(False, False)
            self.axes_list.append(plot)

        self.on_click(self.axes_list[0])(None)


    def set_current_plot(self,plot):

        if self.current_plot != plot:
            self.current_view_box = plot.getViewBox()
            self.current_plot=plot
            if self.tool_bar is not None:
                self.tool_bar.reset()
            return True
        return False
    def on_click(self,plot):
        """
        子图点击  放大
        :param event:
        :return:
        """


        def handler(event):
            # 清除现有布局
            self.ci.clear()

            # 将被双击的子图放在第一行
            self.addItem(plot, row=0, col=0, colspan=4 )

            self.set_current_plot(plot)

            # 将其他子图放在第二行
            other_plots = [p for p in self.axes_list if p != plot]
            for i, other_plot in enumerate(other_plots):
                self.addItem(other_plot, row=1, col=i)

            for col, factor in enumerate([3, 1]):
                self.ci.layout.setRowStretchFactor(col, factor)

        return handler

class CustomGraphicsLayoutWidget(LayoutPlotBase,GraphicsLayoutWidget):

    def __init__(self,*args,**kwargs):
        super().__init__(self )

        GraphicsLayoutWidget.__init__(self,*args,**kwargs)





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/20 22:22
# @Author  : 兵
# @email    : 1747193328@qq.com

import numpy as np
from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QPainter
from pyqtgraph import mkPen, ScatterPlotItem, TextItem, ViewBox,PlotDataItem

from .toolbar import NepDisplayGraphicsToolBar
from .canvas import CustomGraphicsLayoutWidget
from .. import MessageManager, Config
from ..custom_widget.dialog import GetIntMessageBox, SparseMessageBox
from ..io import NepTrainResultData
from ..io.select import farthest_point_sampling
from ..types import Brushes
from NepTrainKit import utils


class NepResultGraphicsLayoutWidget(CustomGraphicsLayoutWidget):
    structureIndexChanged=Signal(int)
    def __init__(self,parent=None):
        super().__init__(parent)
        self._parent=parent
        self.dataset=None
        self.draw_mode=False
        self.setRenderHint(QPainter.Antialiasing, False)
        self.setViewportUpdateMode(self.ViewportUpdateMode.BoundingRectViewportUpdate)

    def set_tool_bar(self, tool):
        self.tool_bar: NepDisplayGraphicsToolBar = tool
        self.tool_bar.panSignal.connect(self.pan)
        self.tool_bar.resetSignal.connect(self.auto_range)
        self.tool_bar.deleteSignal.connect(self.delete)
        self.tool_bar.revokeSignal.connect(self.revoke)
        self.tool_bar.findMaxSignal.connect(self.find_max_error_point)
        self.tool_bar.sparseSignal.connect(self.sparse_point)
        self.tool_bar.penSignal.connect(self.pen)

    def auto_range(self):
        if self.current_plot:
            self.current_plot.getViewBox().autoRange()

    def pan(self, checked):

        if self.current_plot:

            self.current_plot.setMouseEnabled(checked, checked)
            self.current_plot.getViewBox().setMouseMode(ViewBox.PanMode)
    def pen(self, checked):
        if self.current_plot is None:

            return False

        if checked:
            self.draw_mode=True
            # 初始化鼠标状态和轨迹数据
            self.is_drawing = False
            self.x_data = []
            self.y_data = []

        else:
            self.draw_mode=False
            pass
    def mousePressEvent(self, event):
        if not self.draw_mode:
            return super().mousePressEvent(event)

        if event.button() == Qt.MouseButton.LeftButton or event.button() == Qt.MouseButton.RightButton:
            self.is_drawing = True
            self.x_data.clear()  # 清空之前的轨迹数据
            self.y_data.clear()  # 清空之前的轨迹数据
            self.curve = self.current_plot.plot([], [], pen='r')

            self.curve.setData([], [])  # 清空绘制线条，避免对角线


    def mouseReleaseEvent(self, event):

        if not self.draw_mode:
            return super().mouseReleaseEvent(event)
        if event.button() == Qt.MouseButton.LeftButton  or event.button() == Qt.MouseButton.RightButton:
            self.is_drawing = False
            reverse=event.button() == Qt.MouseButton.RightButton
            self.current_plot.removeItem(self.curve)
            # 创建鼠标轨迹的多边形
            if len(self.x_data)>2:

                self.select_point_from_polygon(np.column_stack((self.x_data, self.y_data)),reverse)
            else:
                # 右键的话  选中单个点
                pass
                pos = event.pos()
                mouse_point = self.current_view_box.mapSceneToView(pos)

                x = mouse_point.x()
                self.select_point(mouse_point,reverse)
            return


    def mouseMoveEvent(self, event):
        if not self.draw_mode:
            return super().mouseMoveEvent(event)

        if self.is_drawing:
            pos = event.pos()
            if self.current_plot.sceneBoundingRect().contains(pos):
                # 将场景坐标转换为视图坐标
                mouse_point =self.current_view_box.mapSceneToView(pos)
                x, y = mouse_point.x(), mouse_point.y()
                # 记录轨迹数据
                self.x_data.append(x)
                self.y_data.append(y)

                # 更新绘图
                self.curve.setData(self.x_data, self.y_data)

    def find_max_error_point(self):
        dataset = self.get_current_dataset()
        if dataset is None:
            return
        box= GetIntMessageBox(self._parent,"Please enter an integer N, it will find the top N structures with the largest errors")
        n = Config.getint("widget","max_error_value",10)
        box.intSpinBox.setValue(n)

        if not box.exec():
            return
        nmax= box.intSpinBox.value()
        Config.set("widget","max_error_value",nmax)
        index= (dataset.get_max_error_index(nmax))

        self.select_index(index,False)

    def sparse_point(self):
        if  self.dataset is None:
            return
        box= SparseMessageBox(self._parent,"Please specify the maximum number of structures and minimum distance")
        n_samples = Config.getint("widget","sparse_num_value",10)
        distance = Config.getfloat("widget","sparse_distance_value",0.01)

        box.intSpinBox.setValue(n_samples)
        box.doubleSpinBox.setValue(distance)

        if not box.exec():
            return
        n_samples= box.intSpinBox.value()
        distance= box.doubleSpinBox.value()

        Config.set("widget","sparse_num_value",n_samples)
        Config.set("widget","sparse_distance_value",distance)

        dataset = self.dataset.descriptor
        indices_to_remove = farthest_point_sampling(dataset.now_data,n_samples=n_samples,min_dist=distance)

        # 获取所有索引（从 0 到 len(arr)-1）
        all_indices = np.arange(dataset.now_data.shape[0])

        # 使用 setdiff1d 获取不在 indices_to_remove 中的索引
        remaining_indices = np.setdiff1d(all_indices, indices_to_remove)
        structures = dataset.group_array[remaining_indices]
        self.select_index(structures.tolist(),False)
    def set_dataset(self,dataset):
        self.dataset:NepTrainResultData=dataset
        self.subplot(2,3)
        self.plot_all()

    def convert_pos(self,plot,pos):
        view_range = plot.viewRange()
        x_range = view_range[0]  # x轴范围 [xmin, xmax]
        y_range = view_range[1]  # y轴范围 [ymin, ymax]

        # 将百分比位置转换为坐标
        x_percent = pos[0] # 50% 对应 x 轴中间
        y_percent =  pos[1]  # 20% 对应 y 轴上的某个位置

        x_pos = x_range[0] + x_percent * (x_range[1] - x_range[0])  # 根据百分比计算实际位置
        y_pos = y_range[0] + y_percent * (y_range[1] - y_range[0])  # 根据百分比计算实际位置
        return x_pos,y_pos

    def get_current_dataset(self):
        if self.current_plot is None:
            return None
        plot_index = self.axes_list.index(self.current_plot)
        return self.dataset.dataset[plot_index]
    @utils.timeit
    def plot_all(self):
        self.dataset.select_index.clear()
        _pen = mkPen(None)
        # import time
        # start = time.time()
        for index,_dataset in enumerate(self.dataset.dataset):
            plot=self.axes_list[index]
            plot.disableAutoRange()
            plot.clear()
            if _dataset.title not in ["descriptor"]:
                plot.addLine(angle=45, pos=(0.5, 0.5), pen=mkPen('r', width=2))

            # else:
            plot.setTitle(_dataset.title)

            scatter = ScatterPlotItem(_dataset.x,_dataset.y,data=_dataset.structure_index,
                                      brush=Brushes.TransparentBrush ,pen=mkPen(color="blue", width=0.5),
                                      symbol='o',size=7,

                                      )


            scatter.sigClicked.connect(self.item_clicked)
            plot.scatter=scatter
            # plot.setDownsampling(auto=True, mode='peak')

            plot.addItem(scatter)

            # 优化抗锯齿设置

            # 设置视图框更新模式
            plot.autoRange()
            if _dataset.title not in ["descriptor"]:

                pos=self.convert_pos(plot,(0 ,1))
                text=f"rmse: {_dataset.get_formart_rmse()}"
                text_item = TextItem(text=text ,color=(231, 63, 50))
                text_item.setPos(*pos)
                plot.addItem(text_item)

        # print(time.time()-start)
        #5.67748498916626
    def item_clicked(self,scatter_item,items,event):

        if items.any():
            item=items[0]

            self.structureIndexChanged.emit(item.data())

    def delete(self):
        if self.dataset is not None and self.dataset.select_index:

            self.dataset.delete_selected()
            self.plot_all()


    def select_point_from_polygon(self,polygon_xy,reverse ):
        index=self.is_point_in_polygon(np.column_stack([self.current_plot.scatter.data["x"],self.current_plot.scatter.data["y"]]),polygon_xy)
        index = np.where(index)[0]
        select_index=self.current_plot.scatter.data[index]["data"].tolist()
        self.select_index(select_index,reverse)


    def select_point(self,pos,reverse):
        items=self.current_plot.scatter.pointsAt(pos)
        if len(items):

            item=items[0]

            index=item.index()
            structure_index =item.data()
            self.select_index(structure_index,reverse)

    def select_index(self,structure_index,reverse):
        if isinstance(structure_index,int):
            structure_index=[structure_index]
        elif isinstance(structure_index,np.ndarray):
            structure_index=structure_index.tolist()

        if not structure_index:
            return
        if reverse:
            self.dataset.uncheck(structure_index)
            self.update_axes_color(structure_index, Brushes.TransparentBrush)

        else:

            self.dataset.select(structure_index)

            self.update_axes_color(structure_index, Brushes.RedBrush)


    def update_axes_color(self,structure_index,color=Brushes.RedBrush):


        for i,plot in enumerate(self.axes_list):

            if not hasattr(plot,"scatter"):
                continue
            structure_index_set= set(structure_index)
            index_list = [i for i, val in enumerate(plot.scatter.data["data"]) if val in structure_index_set]

            plot.scatter.data["brush"][index_list]=   color
            plot.scatter.data['sourceRect'][index_list] = (0, 0, 0, 0)


            plot.scatter.updateSpots(  )





    def revoke(self):
        """
        如果有删除的结构  撤销上一次删除的
        :return:
        """

        if self.dataset and  self.dataset.is_revoke:
            self.dataset.revoke()
            self.plot_all()

        else:
            MessageManager.send_info_message("No undoable deletion!")
